/*Utility routines for poly types*/

#include "local.h"
#include <stdio.h>
#include <math.h>
#include "son.h"
#include "em.typ"

#define POINT_PAGE_SIZE 16
#define POLY_PAGE_SIZE 8
Rpoly *add_rpoly();
Ipoly *add_ipoly();

static FILE *tracefile = NULL;
static int debug_poly = FALSE;


clear_rpoly(poly) Rpoly *poly;
/*************************************************************/
/* Close down all allocated memory in poly.                  */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
{
  if (poly->xy0!=NULL) sFree((char *)poly->xy0);
  poly->xy0 = NULL;
  poly->n = 0;
  poly->max = 0;
  poly->mode = 0;
  poly->icond = -1;
  poly->xmin = poly->ymin = poly->xmax = poly->ymax = 0;
  poly->manhatten = -1;
  poly->id = 0;
  poly->xgeom_id = 0;
} /*clear_rpoly*/

clear_ipoly(poly) Ipoly *poly;
/*************************************************************/
/* Close down all allocated memory in poly.                  */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
{
  if (poly->xy0!=NULL) sFree((char *)poly->xy0);
  poly->xy0 = NULL;
  poly->n = 0;
  poly->max = 0;
} /*clear_ipoly*/

clear_poly_list(list) Poly_list *list;
/*************************************************************/
/* Close down all allocated memory in list.                  */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
{ int i;

  if (list->rpoly0 != NULL)
  { for (i=0; i<list->n; i++) clear_rpoly(list->rpoly0+i);
    sFree((char *)list->rpoly0);
  }

  if (list->ipoly0 != NULL)
  { for (i=0; i<list->n; i++) clear_ipoly(list->ipoly0+i);
    sFree((char *)list->ipoly0);
  }

  list->rpoly0 = NULL;  list->ipoly0 = NULL;
  list->n = 0;  list->max = 0;

} /*clear_poly_list*/

clear_cir(cir)  Cir *cir;
/*************************************************************/
/* Close down all allocated memory in cir.                   */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
{ msub_list *lev;
  wt_info *wt,**wt0;
  int i;

  if (cir->plev0)
  { for (i=0; i<cir->n; i++) clear_poly_list(cir->plev0+i);
    sFree( (char *)cir->plev0);  cir->plev0 = NULL;
  }

  if (cir->lev0)
  { for (lev=cir->lev0; lev<cir->lev0+cir->n; lev++) clear_lev(lev);
    sFree( (char *)cir->lev0 );  cir->lev0 = NULL;
  }

  cir->n = 0;

  if (cir->ports)
  { clear_ports(cir->ports);  sFree( (char *)cir->ports);  cir->ports = NULL; }
  if (cir->vias)
  { clear_vias(cir->vias);  sFree((char *)cir->vias);  cir->vias = NULL; }

  /*Clear out all wieght info*/
  if (cir->wt0)
  { for (wt0=cir->wt0; wt0<cir->wt0+NUM_ESUB*cir->nloc; wt0++) if (*wt0!=NULL)
    { wt = *wt0;  sFree((char *)wt->wx);  sFree((char *)wt->wy);  sFree((char *)wt); }
    sFree((char *)cir->wt0);  cir->wt0 = NULL;
  }
  cir->nloc = 0;

} /*clear_cir*/

init_rpoly(poly) Rpoly *poly;
/*************************************************************/
/* Initialize all info in poly.                              */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
{
  poly->n = 0;
  poly->max = 0;
  poly->xy0 = NULL;
  poly->mode = 0;
  poly->icond = -1;
  poly->xmin = 1;
  poly->ymin = 1;
  poly->xmax = 100;
  poly->ymax = 100;
  poly->manhatten = -1;
  poly->unsel = 0;
  poly->id = 0;
} /*init_rpoly*/

init_ipoly(poly) Ipoly *poly;
/*************************************************************/
/* Initialize all info in poly.                              */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
{
  poly->n = 0;
  poly->max = 0;
  poly->xy0 = NULL;
} /*init_ipoly*/

init_poly_list(list) Poly_list *list;
/*************************************************************/
/* Initialize all info in list.                              */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
{
  list->n = 0;
  list->max = 0;
  list->rpoly0 = NULL;
  list->ipoly0 = NULL;
} /*init_poly_list*/

init_cir(cir,nlev) Cir *cir;  int nlev;
/*************************************************************/
/* Initialize all info in cir.                               */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
/* jcr20mar90 Modified to add Via to gnd only level, lev nlev.*/
{ int i;
  nlev++;  /*Extra lev for vias to gnd*/
  cir->n = nlev;

  cir->plev0 = (Poly_list *)new(nlev*sizeof(Poly_list));
  for (i=0; i<nlev; i++) init_poly_list(cir->plev0+i);

  cir->lev0 = (msub_list *)new(nlev*sizeof(msub_list));
  for (i=0; i<nlev; i++) init_lev( cir->lev0+i );

  cir->nloc = 0;
  cir->max_wt_len = 0;
  cir->wt0 = NULL;

  cir->ports = (port_list *)new(sizeof(port_list));
  init_ports(cir->ports);

  cir->vias = (via_list *)new(sizeof(via_list));
  init_vias(cir->vias);

} /*init_cir*/

add_rpoly_pt(poly,x,y)  Rpoly *poly;  Rtyp x,y;
/*************************************************************/
/* Append a single point to plist.                           */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
{
  if (poly->n >= poly->max)
  { poly->max += POINT_PAGE_SIZE;
    poly->xy0 = ( poly->xy0==NULL )
       ? (Rtyp *)new( 2*poly->max*sizeof(Rtyp) )
       : (Rtyp *)rnew( (char *)poly->xy0, 2*poly->max*sizeof(Rtyp) );
  }
 
  *(poly->xy0+2*poly->n) = x;
  *(poly->xy0+2*poly->n+1) = y;
  poly->unsel++;
  poly->n++;
} /*add_rpoly_pt*/

add_ipoly_pt(poly,x,y)  Ipoly *poly;  int x,y;
/*************************************************************/
/* Append a single point to plist.                           */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
{ int *xp;

  if (poly->n >= poly->max)
  { poly->max += POINT_PAGE_SIZE;
    poly->xy0 = ( poly->xy0==NULL )
       ? (int *)new( 2*poly->max*sizeof(int) )
       : (int *)rnew( (char *)poly->xy0, 2*poly->max*sizeof(int) );
  }

  xp = poly->xy0 + 2 * poly->n;
  /*Don't repeat same point*/
  if (xp>poly->xy0 && *(xp-2) == x && *(xp-1) == y) return;
 
  *xp = x;
  *(xp + 1) = y;
  poly->n++;
} /*add_ipoly_pt*/

ins_ipoly_pt(poly,x,y,cmp)  Ipoly *poly;  int x,y;  int (*cmp)();
/*************************************************************/
/* Insert a single point in the sorted plist.                */
/* (C) COPYRIGHT 1990 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr10jul90 */ 
{ int *ixy,*xy0,*xy1;
  int len,dif,xy[2];

  if (poly->n >= poly->max)
  { poly->max += POINT_PAGE_SIZE;
    poly->xy0 = ( poly->xy0==NULL )
       ? (int *)new( 2*poly->max*sizeof(int) )
       : (int *)rnew( (char *)poly->xy0, 2*poly->max*sizeof(int) );
  }

  xy[0] = x;  xy[1] = y;  /*Find insertion pt for this value*/
  xy0 = poly->xy0;
  xy1 = xy0 + 2 * poly->n;
  /*Put pt exceeding last pt + new value at end*/
  *xy1 = abs(x) + *(xy1-2);  *(xy1+1) = abs(y) + *(xy1-1);

  /*Find the insertion point, while len!=0*/
  while ( len = (int)(xy1 - xy0) )
  { ixy = xy0 + ((len>>1) & ~1);
    dif = cmp(ixy,xy);
    if (dif>0) xy1 = ixy;
    else if (dif<0)
    { if (xy0+2 == xy1) xy0 = xy1;  /*Done*/
      else xy0 = ixy;
    }
    else return;  /*Don't insert dupe*/
  }

  /*Move everything up one*/
  for (ixy=poly->xy0+2*poly->n-1; ixy>=xy0; ixy--) *(ixy+2) = *ixy;
  /*Insert point*/
  *xy0 = x;  *(xy0+1) = y;
  poly->n++;
} /*ins_ipoly_pt*/

insert_ipoly_pt(ipoly,ix,iy,ipt)  Ipoly *ipoly; int ix,iy;  int ipt;
/******************************************************************/
/* Insert the given point before the ipt-th point.                */
/* Copyright (c) 1989 Sonnet Software, Inc., All rights reserved. */
/******************************************************************/
/* jcr19dec90 */
{ int *ixy0,*ixy0_stop,*iixy0;

  if (ipt <= 0 || ipt > ipoly->n)
    announce_bug("insert_ipoly_pt");

  /*Make room for the point*/
  add_ipoly_pt(ipoly,0,0);

  /*Now move everything up one*/
  ixy0_stop = ipoly->xy0 + 2 * ipt;
  ixy0 = ipoly->xy0 + 2 * ipoly->n - 1;
  for (iixy0=ixy0; iixy0>ixy0_stop;  iixy0--) {
  	*iixy0 = *(iixy0-2);
  }

  *ixy0_stop = ix;  
  *(ixy0_stop+1) = iy;

} /*insert_ipoly_pt*/

Rpoly *add_rpoly(plist,xyp,n) Poly_list *plist;  Rtyp *xyp;  int n;
/*************************************************************/
/* Append polygon to end of plist.                           */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
{ Rpoly *rpoly;
  int i;
  static xgeom_id = 1;

  if (plist->n >= plist->max)
  { plist->max += POLY_PAGE_SIZE;
    plist->rpoly0 = ( plist->rpoly0==NULL )
       ? (Rpoly *)new( plist->max*sizeof(Rpoly) )
       : (Rpoly *)rnew( (char *)plist->rpoly0, plist->max*sizeof(Rpoly) );
    plist->ipoly0 = ( plist->ipoly0==NULL )
       ? (Ipoly *)new( plist->max*sizeof(Ipoly) )
       : (Ipoly *)rnew( (char *)plist->ipoly0, plist->max*sizeof(Ipoly) );
  }

  init_rpoly( rpoly = plist->rpoly0+plist->n );
  init_ipoly( plist->ipoly0 + (plist->n++) );
  rpoly->xgeom_id = xgeom_id++;

  for (i=0; i<n; i++) add_rpoly_pt(rpoly,*(xyp+2*i),*(xyp+2*i+1));

  return(rpoly);

} /*add_rpoly*/

Ipoly *add_ipoly(plist,xyp,n) Poly_list *plist;  int *xyp;  int n;
/*************************************************************/
/* Append polygon to end of plist.                           */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
{ Ipoly *ipoly;
  int i;

  if (plist->n >= plist->max)
  { plist->max += POLY_PAGE_SIZE;
    plist->rpoly0 = ( plist->rpoly0==NULL )
       ? (Rpoly *)new( plist->max*sizeof(Rpoly) )
       : (Rpoly *)rnew( (char *)plist->rpoly0, plist->max*sizeof(Rpoly) );
    plist->ipoly0 = ( plist->ipoly0==NULL )
       ? (Ipoly *)new( plist->max*sizeof(Ipoly) )
       : (Ipoly *)rnew( (char *)plist->ipoly0, plist->max*sizeof(Ipoly) );
  }

  init_rpoly( plist->rpoly0+plist->n );
  init_ipoly( ipoly = plist->ipoly0 + (plist->n++) );

  for (i=0; i<n; i++) add_ipoly_pt(ipoly,*(xyp+2*i),*(xyp+2*i+1));

  return(ipoly);

} /*add_ipoly*/

remove_poly(plist,i) Poly_list *plist;  int i;
/*************************************************************/
/* Remove ith polygon from list.                             */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr14may89 */ 
{ Ipoly *ipoly;
  Rpoly *rpoly;

  if ( i >= plist->n || i < 0 )
    announce_bug("remove_poly");

  clear_ipoly(plist->ipoly0+i);
  for (ipoly=plist->ipoly0+i; ipoly+1<plist->ipoly0+plist->n; ipoly++)
    move_ipoly(ipoly,ipoly+1);

  clear_rpoly(plist->rpoly0+i);
  for (rpoly=plist->rpoly0+i; rpoly+1<plist->rpoly0+plist->n; rpoly++)
    move_rpoly(rpoly,rpoly+1);

  plist->n--;

} /*remove_poly*/

move_plev( to, from )
/******************************************************************/
/* Move cir->plev0 from 'from' to 'to'				  */
/* Copyright (c) 1989 Sonnet Software, Inc., All rights reserved. */
/******************************************************************/
/* krs09Apr90 */
Poly_list *to, *from;
{
	to->n       = from->n;
	to->max     = from->max;
	to->rpoly0  = from->rpoly0;
	to->ipoly0  = from->ipoly0;
}

move_ipoly(to,from)  Ipoly *to,*from;
/*************************************************************/
/* Do the indicated move.                                    */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr14may89 */ 
{
  if (from==to) return;
  if (from==NULL || to==NULL)
    announce_bug("move_ipoly");
  clear_ipoly(to);
  to->n = from->n;  from->n = 0;
  to->max = from->max;  from->max = 0;
  to->xy0 = from->xy0;  from->xy0 = NULL;
} /*move_ipoly*/

move_rpoly(to,from)  Rpoly *to,*from;
/*************************************************************/
/* Do the indicated move.                                    */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr14may89 */ 
{ 
  if (from==to) return;
  if (from==NULL || to==NULL)
    announce_bug("move_rpoly");
  clear_rpoly(to);
  to->n = from->n;  from->n = 0;
  to->max = from->max;  from->max = 0;
  to->xy0 = from->xy0;  from->xy0 = NULL;
  to->mode = from->mode;  from->mode = 0;
  to->icond = from->icond;  from->icond = -1;
  to->xmin = from->xmin;  from->xmin = 0;
  to->ymin = from->ymin;  from->ymin = 0;
  to->xmax = from->xmax;  from->xmax = 0;
  to->ymax = from->ymax;  from->ymax = 0;
  to->manhatten = from->manhatten;
  to->unsel = from->unsel;  from->unsel = 0;
  to->id = from->id;  from->id = 0;
  to->xgeom_id = from->xgeom_id;  from->xgeom_id = 0;

} /*move_rpoly*/

pt_in_rpoly(poly,x,y) Rpoly *poly;  Rtyp x,y;
/*************************************************************/
/* Return 1 if pt in poly, o.w. 0.                           */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
{ Rtyp *x0,*x1,*y0,*y1,*xstop;
  int cnt;

  if (poly->n<3) return(0);

  cnt = 0;
  xstop = poly->xy0 + 2 * poly->n;
  for ( y1=1+(x1=1+(y0=1+(x0=poly->xy0))); x1<xstop; x0=x1,y0=y1,y1++,x1=y1++)
    if ( ((*x0<=x) ^ (*x1<=x)) && (*y0>=y || *y1>=y) )
    { if (*y0>=y && *y1>=y) cnt++;
      else cnt += (*x1 > *x0)
		  ? ( (x - *x0)*(*y1 - *y0) >=  (y - *y0)*(*x1 - *x0) )
		  : ( (x - *x0)*(*y1 - *y0) <   (y - *y0)*(*x1 - *x0) );
    }
  return(cnt&1);
} /*pt_in_rpoly*/

pt_in_ipoly(poly,x,y) Ipoly *poly;  int x,y;
/*************************************************************/
/* Return 1 if pt in poly, o.w. 0.                           */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
{ int *x0,*x1,*y0,*y1,*xstop;
  int cnt;

  if (poly->n<3) return(0);

  cnt = 0;
  xstop = poly->xy0 + 2 * poly->n;
  for ( y1=1+(x1=1+(y0=1+(x0=poly->xy0))); x1<xstop; x0=x1,y0=y1,y1++,x1=y1++)
    if ( ((*x0<=x) ^ (*x1<=x)) && (*y0>y || *y1>y) )
    { if (*y0>=y && *y1>=y) cnt++;
      else cnt += (*x1 > *x0)
		  ? ( (x - *x0)*(*y1 - *y0) >=  (y - *y0)*(*x1 - *x0) )
		  : ( (x - *x0)*(*y1 - *y0) <   (y - *y0)*(*x1 - *x0) );
    }
  return(cnt&1);
} /*pt_in_ipoly*/

pt_in_dipoly(poly,x,y) Ipoly *poly;  int x,y;
/*************************************************************/
/* Return 1 if pt in on a diagonal edge (removed), o.w. 0.   */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr16mar92 */ 
{ int *x0,*x1,*y0,*y1,*xstop;

  if (poly->n<3) return(0);

  xstop = poly->xy0 + 2 * poly->n;
  
  /*Skip middle point and look for x,y on resulting diagonal*/
  x0 = poly->xy0;
  y0 = poly->xy0+1;
  x1 = poly->xy0+4;
  y1 = poly->xy0+5;

  for ( ; x1<xstop; x0 += 2, y0 += 2, x1 += 2, y1 += 2)
    if ( ((*x0<=x) ^ (*x1<=x)) && (*y0>y || *y1>y) &&
            (x - *x0)*(*y1 - *y0) == (*x1 - *x0)*(y - *y0) ) return(1); /*Pt on edge*/

  /*Check for poss. removed diagonal at end and beg. of poly*/
  x0 = poly->xy0+2;
  y0 = poly->xy0+3;
  x1 = xstop-4;
  y1 = xstop-3;
  if ( ((*x0<=x) ^ (*x1<=x)) && (*y0>y || *y1>y) &&
          (x - *x0)*(*y1 - *y0) == (*x1 - *x0)*(y - *y0) ) return(1); /*Pt on edge*/
  return(0);
} /*pt_in_dipoly*/

ck_manhatten(rpoly,Box)  Rpoly *rpoly;  Box_info *Box;
/*************************************************************/
/* See if rpoly has nothing but vert and horz edges.         */
/* (C) COPYRIGHT 1993 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr31mar93 */
{ int ix0,iy0,ix1,iy1;
  Rtyp *xp,*yp,*xstop;
  Rtyp rx1,ry1,dx,dy;

  if ( rpoly->manhatten >= 0 )
    return( rpoly->manhatten );

  rpoly->manhatten = 0;

  if (rpoly->n < 5) return( rpoly->manhatten );

  dx = Box->A / Box->Capm;  dy = Box->B / Box->Capn;

  xp = rpoly->xy0;
  yp = xp + 1;
  rx1 = *xp / dx;  ry1 = *yp / dy;
  ix1 = intx(rx1);  iy1 = inty(ry1);

  /*We ck 2 pts at a time, so stop with 1st pt one short of end*/
  xstop = rpoly->xy0 + 2 * (rpoly->n - 1);

  for (yp++,xp=yp++; xp<xstop; yp++,xp=yp++)
  { ix0 = ix1;  iy0 = iy1;
    rx1 = *xp/dx;  ry1 = *yp/dy;
    ix1 = intx(rx1);  iy1 = inty(ry1);
    if ( ix0 != ix1 && iy0 != iy1 )
      return( rpoly->manhatten );  /*i.e., return(0) */
  }

  rpoly->manhatten = 1;
  return( rpoly->manhatten );
} /*ck_manhatten*/

poly_rtoi(rpoly,ipoly,dx,dy)  Rpoly *rpoly;  Ipoly *ipoly;  Rtyp dx,dy;
/*************************************************************/
/* Set up ipoly to correspond to rpoly.                      */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr28apr89 */ 
{ Rtyp *x,*y,*xstop;
  int ix0,iy0,ix1,iy1,ix,iy;
  int sgnx,sgny,delx,dely;
  Rtyp rx0,ry0,rx1,ry1,m;
  int *ixp,first_pt,mode;
  int old_dist,new_dist;
  Rtyp xint,yint,dist;

  if ( rpoly==NULL || ipoly==NULL || dx<=0.0 || dy<=0.0 ||
       (rpoly->mode!=STAIRCASE && rpoly->mode!=DIAGONAL && rpoly->mode!=SIMPLE) )
    announce_bug("poly_rtoi");

  mode = rpoly->mode;
  dist = (mode==DIAGONAL) ? 2.0 : 1.0;
  x = rpoly->xy0 + rpoly->n - 4;  y = x + 1;  /*Next to last x,y pair*/
  rx1 = *x/dx;      ry1 = *y/dy;
  ix1 = intx(rx1);  iy1 = inty(ry1);
  clear_ipoly(ipoly);
  init_ipoly(ipoly);
  if (rpoly->n <= 0) return;
  xstop = rpoly->xy0 + 2 * rpoly->n;
  first_pt = 1;
  for (x=rpoly->xy0,y=x+1; x<xstop; y++,x=y++) while (1) /*continue until break*/
  { rx0 = rx1;        ry0 = ry1;
    rx1 = *x/dx;      ry1 = *y/dy;
    ix0 = ix1;        iy0 = iy1;
    ix1 = intx(rx1);  iy1 = inty(ry1);
    if ( mode==SIMPLE || ix0==ix1 || iy0==iy1 || (mode==DIAGONAL && abs(ix0-ix1)==abs(iy0-iy1)) || first_pt )
    { /*In very rare cases, the line segment from the current last point in ipoly
        to ix1,iy1 is not hor. or vert, (or 45 deg diag) take care of that case */
      if ( !first_pt )
      { ixp = ipoly->xy0 + 2 * ipoly->n - 2;
        sgnx = (*ixp>ix1) ? -1 : 1;  delx = abs(ix1 - *ixp);
        sgny = (*(ixp+1)>iy1) ? -1 : 1;  dely = abs(iy1 - *(ixp+1));
        if ( !first_pt && delx && dely && ( mode!=DIAGONAL || delx!=dely ) )
        { if ( delx==2 ) add_ipoly_pt(ipoly,*ixp,iy1);
          else if ( dely==2 ) add_ipoly_pt(ipoly,ix1,*(ixp+1));
          else if ( abs(delx-dely)==2 ) /*diagonal off by one cell*/
          { if (delx>dely) delx = dely;  /*Select shortest true diagonal line*/
            add_ipoly_pt(ipoly,*ixp+sgnx*delx,*(ixp+1)+sgny*delx);
          }
          else {
	    if ( debug_poly )
	      dump_rpoly(rpoly,dx,dy);
	    announce_bug("poly_rtoi(0)");
	  }
        }
      }
      add_ipoly_pt(ipoly,ix1,iy1);  first_pt = 0;  break;
    }

    /*divide by zero possibility eliminated by above if statement*/
    m = (ry1 - ry0) / (rx1 - rx0);
    new_dist = abs(ix0-ix1) + abs(iy0-iy1);
    old_dist = new_dist + 1;
    ix = ix0;  iy = iy0;
    while ( old_dist > new_dist )
    { old_dist = new_dist;
      /*Calculate some intercepts*/
      yint = (ix0>ix1) ? (ry0 - iy) - m * (dist + rx0 - ix)
		       : (ry0 - iy) + m * (dist - rx0 + ix);
      xint = (iy0>iy1) ? (rx0 - ix) - (dist + ry0 - iy) / m
		       : (rx0 - ix) + (dist - ry0 + iy) / m;
      /*If intercept is in wrong direction, set to zero*/
      if ( yint>0.0 && iy1<iy0 ) yint = 0.0;
      if ( yint<0.0 && iy1>iy0 ) yint = 0.0;
      if ( xint>0.0 && ix1<ix0 ) xint = 0.0;
      if ( xint<0.0 && ix1>ix0 ) xint = 0.0;
      if (mode==STAIRCASE)
      { if ( fabs(fabs(yint)-fabs(xint)) < TINY ) /*Diag line, always go on low x,y side*/
	{ if (ix0<ix1 && iy0<iy1) /* :\ */
	    ix += 2;
	  else if (ix0<ix1 && iy0>iy1) /* :/ */
	    iy -= 2;
	  else if (ix0>ix1 && iy0>iy1) /* \: */
	    iy -= 2;
	  else if (ix0>ix1 && iy0<iy1)  /* /: */
	    ix -= 2;
        }
        else if ( fabs(yint)<=fabs(xint) )
	     ix += (ix0<ix1) ? 2 : -2;
	else
	     iy += (iy0<iy1) ? 2 : -2;
      }
      if (mode==DIAGONAL)
      { if ( fabs(yint)>1.0 )
	     iy += (iy0<iy1) ? 2 : -2;
        if ( fabs(xint)>1.0 )
	     ix += (ix0<ix1) ? 2 : -2;
      }
      new_dist = abs(ix-ix1) + abs(iy-iy1);
      if ( old_dist > new_dist )
      { /*If new point is on a straight line with previous points,
         *just update last point*/
        if (ipoly->n<=1) add_ipoly_pt(ipoly,ix,iy);
        else
        { ixp = ipoly->xy0 + 2 * ipoly->n - 4;
          if ( (*ixp - *(ixp+2)) * (*(ixp+3) - iy) == (*(ixp+1) - *(ixp+3)) * (*(ixp+2) - ix) )
  	    { *(ixp+2) = ix;  *(ixp+3) = iy; }
  	  if ( ( *ixp == *(ixp+2) && *ixp == ix ) || ( *(ixp+1) == *(ixp+3) && *(ixp+1) == iy ) )
	    { *(ixp+2) = ix;  *(ixp+3) = iy; }
          else add_ipoly_pt(ipoly,ix,iy);
        }
      }
    }
    break;  /*break while(1), do next index in for*/
  }

  /*Make sure the last point is the same as the first*/
  add_ipoly_pt(ipoly,*ipoly->xy0,*(ipoly->xy0+1));

  /*filter out acute internal angles*/
  ipoly_rm_acute(ipoly);

} /*poly_rtoi*/

add_vias(cir,Box)  Cir *cir;  Box_info *Box;
/*************************************************************/
/* Add Vias to circuit.                                      */
/* (C) COPYRIGHT 1990 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr26mar90 */ 
{ via_typ *via,*via_stop;
  int *vxy,*ivxy,*ivxy_stop;
  int num;
  Rtyp dx,dy,*x0p;
  Poly_list *plev;
  Rpoly *rpoly,via_poly_struct,*via_poly;
  msub_list *lev;

  if (cir->vias->n <= 0) return;

  via_poly = &via_poly_struct;
  init_rpoly(via_poly);
  dx = Box->A / Box->Capm;
  dy = Box->B / Box->Capn;
  via = cir->vias->via0;
  via_stop = via + cir->vias->n;

  for ( ; via<via_stop; via++)
  { plev = cir->plev0 + via->ilev;
    lev = cir->lev0 + via->ilev;
    rpoly = plev->rpoly0 + via->poly_index;
    x0p = rpoly->xy0 + 2 * via->ipt;
    via2vxy(rpoly,x0p,dx,dy,via->typ,&vxy,&num);
    /*Add in each via*/
    if (num>0)
    { ivxy_stop = vxy+2*num;
      for (ivxy=vxy; ivxy<ivxy_stop; ivxy+=2)
        if (*ivxy>0 && *ivxy<Box->Capm && *(ivxy+1)>0 && *(ivxy+1)<Box->Capn )
          add_msub(MVIA,ivxy,Box,lev);
      /*Make rpoly for base of via and add in*/
      vxy2rpoly(vxy,num,via_poly,dx,dy);
      via_poly->icond = rpoly->icond;
      /*Always add via_poly to via level so via is sure to get icond*/
      rpoly = add_rpoly(plev,via_poly->xy0,via_poly->n);
      rpoly->icond = via_poly->icond;
      /*Do not add to plev -1, which doesn't exist*/
      if ( via->ilev>0 ) /*We want subsectioning done properly on top end of via*/
      { rpoly = add_rpoly(plev-1,via_poly->xy0,via_poly->n);
        rpoly->icond = via_poly->icond;
      }
    }
  }
  via2vxy( (Rpoly *)NULL,(Rtyp *)NULL,0.0,0.0,0,&vxy,&num);  /*Clear out mem*/
  clear_rpoly(via_poly);
} /*add_vias*/

via2vxy(poly,x0p,dx,dy,typ,vxyp,nump)  Rpoly *poly;  Rtyp *x0p,dx,dy;
				       int typ,**vxyp,*nump;
/*************************************************************/
/* Convert end pts (*x0p...) to int list of via cen coord.   */
/* (C) COPYRIGHT 1990 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr26mar90 */ 
/* Coord list rtrnd in **vxyp, with 2*(*nump) pts*/
/* jcr10dec92 modified so old and new vias both allowed (VIA,EVIA)*/
{ static int *vxy = NULL;
  static int max = 0;
  int *ivxy,*jvxy,num;
  int ix0,iy0,ix1,iy1,xdiff,ydiff;
  Rtyp rx0,ry0,rx1,ry1,rlen,rdx,rdy,xtest,ytest;
  Rtyp delt=1.0e-5;  /*So int(x/y)_odd always goes same way on edge*/

  /*If x0p is NULL, clear out static memory*/
  /*Return *vxyp, *nump, in case vxy == NULL for a real polygon*/
  if (x0p==NULL)
  { if (vxy!=NULL) sFree((char *)vxy);
    vxy = NULL;  max = 0;  *vxyp = NULL;  *nump = 0;
    return;
  }

  /*Initialize the via center array*/
  num = 0;
  if (vxy==NULL)
  { max = 16;
    vxy = (int *)new( 2*max*sizeof(int) );
  }

  rx0 = *x0p / dx;
  ry0 = *(x0p+1) / dy;
  rx1 = *(x0p+2) / dx;
  ry1 = *(x0p+3) / dy;

  if ( typ == VIA_EDGE )
  { /*Check if vertical line and vias could be on either side*/
    ix0 = (rx0>=0.0) ? intx_odd(rx0+delt) : intx_odd(rx0-delt);
    iy0 = (ry0>=0.0) ? intx_odd(ry0+delt) : intx_odd(ry0-delt);
    /*if vertical line and on even coord edge, move to inside polygon*/
    if ( fabs(rx0-rx1) < 0.01 && fabs( (Rtyp)ix0 - rx0 ) > 0.99 )
    { xtest = 0.5 * ( *x0p + *(x0p+2) + 1.0e-4 * dx );
      ytest = 0.5 * ( *(x0p+1) + *(x0p+3) );
      rx0 = rx1 = ( pt_in_rpoly(poly,xtest,ytest) ) ? rx0 + 1.0 : rx0 - 1.0;
    }
    /*if horizontal line and on even coord edge, move to inside polygon*/
    if ( fabs(ry0-ry1) < 0.01 && fabs( (Rtyp)iy0 - ry0 ) > 0.99 )
    { xtest = 0.5 * ( *x0p + *(x0p+2) );
      ytest = 0.5 * ( *(x0p+1) + *(x0p+3) + 1.0e-4 * dy );
      ry0 = ry1 = ( pt_in_rpoly(poly,xtest,ytest) ) ? ry0 + 1.0 : ry0 - 1.0;
    }
  }

  rlen = hypot( rx1-rx0,ry1-ry0 );
  /*Handle single via line*/
  if (rlen<1.5)
  { rx1 = 0.5*(rx0+rx1);
    ry1 = 0.5*(ry0+ry1);
    *vxy = (rx1>=0.0) ? intx_odd(rx1+delt) : intx_odd(rx1-delt);
    *(vxy+1) = (ry1>=0.0) ? inty_odd(ry1+delt) : inty_odd(ry1-delt);
    *vxyp = vxy;
    *nump = 1;
    return;
  }
  rdx = 2.0 * (rx1-rx0) / rlen;
  rdy = 2.0 * (ry1-ry0) / rlen;

  /*If rdx is pos, reverse direction of line, so
    stopping condx in vxy2rpoly is not passed over*/
  if (rdx>0.0)
  { rdx = -rdx;  rdy = -rdy;
    rx0 = rx1;  ry0 = ry1;
  }

  /*Always add at least one via*/
  rx1 = rx0 + 0.5 * rdx;
  ry1 = ry0 + 0.5 * rdy;
  num = 1;
  *vxy = (rx1>=0.0) ? intx_odd(rx1+delt) : intx_odd(rx1-delt);
  *(vxy+1) = (ry1>=0.0) ? inty_odd(ry1+delt) : inty_odd(ry1-delt);
  rx1 += rdx;  ry1 += rdy;

  while ( rlen > hypot( rx1-rx0,ry1-ry0 ) )
  { if (num==max)
    { max += 16;
      vxy = (int *)rnew((char *)vxy,2*max*sizeof(int));
    }
    ix1 = (rx1>=0.0) ? intx_odd(rx1+delt) : intx_odd(rx1-delt);
    iy1 = (ry1>=0.0) ? inty_odd(ry1+delt) : inty_odd(ry1-delt);

    /*Check for missing via*/
    xdiff = ix1 - *(vxy+2*num-2);
    ydiff = iy1 - *(vxy+2*num-1);
    if ( abs(xdiff)>3 || abs(ydiff)>3 )
    { if (xdiff>=4) xdiff = 2;
      if (ydiff>=4) ydiff = 2;
      *(vxy+2*num) = ix1 - xdiff;
      *(vxy+2*num+1)   = iy1 - ydiff;
      num++;
    }

    /*Filter out dupe vias*/
    if ( *(vxy+2*num-2)!=ix1 || *(vxy+2*num-1)!=iy1 )
    { *(vxy+2*num) = ix1;
      *(vxy+2*num+1)   = iy1;
      num++;
    }
    rx1 += rdx;  ry1 += rdy;
  }

  /*Make sure each via has only one touching via on each side.
    Diagonally spaced vias take precidence*/
  for (ivxy=vxy; ivxy<vxy+2*num-4; ivxy+=2)
  { ix0 = *ivxy - *(ivxy+2);
    ix1 = *ivxy - *(ivxy+4);
    iy0 = *(ivxy+1) - *(ivxy+3);
    iy1 = *(ivxy+1) - *(ivxy+5);
    if ( (abs(ix0)==2 && ix0==ix1 && iy0==0 && abs(iy1)==2 ) ||
         (abs(iy0)==2 && iy0==iy1 && ix0==0 && abs(ix1)==2 ) )
    { for (jvxy=ivxy+2; jvxy<vxy+2*num-2; jvxy++)  /*Remove adj via*/
        *jvxy = *(jvxy+2);
      num--;
    }
  }

  *vxyp = vxy;
  *nump = num;
} /*via2vxy*/

vxy2rpoly(vxy,num,poly,dx,dy) int *vxy,num; Rpoly *poly;
                                  Rtyp dx,dy;
/*************************************************************/
/* Use list of via centers to make enclosing rpoly.          */
/* (C) COPYRIGHT 1990 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr26mar90 */ 
/* This routine assumes NO Dupes in vxy and all vias are touching*/
{ int *cen,*last;
  int ix0,iy0,ix0_old,iy0_old,ix1,iy1;
  int corner,side,inc;

/*    corner 1    side 0    corner 0
         *---------------------*        Y-axis
         |                     |          ^
  side 1 |     via center      | side 3   |
         |                     |          *---> X-axis
         *---------------------*
      corner 2    side 2    corner 3

  Algorithm starts at corner 0, side 0 and cks to see which
  way to go.  Only corner XOR side changed when staying on
  same via, before next check.  Only two vias are allowed
  to be touching any other via, and those two vias may not
  be touching each other.  Also vxy list must not have x
  value increasing as index increases
*/

  clear_rpoly(poly);
  if (num==0) return;

  cen = vxy;
  last = vxy + 2*num - 2;
  corner = 0;
  side = 0;

  /*Do the 0th corner of the 0th side*/
  ix0_old = ix0 = *cen + 1;
  iy0_old = iy0 = *(cen+1) + 1;
  add_rpoly_pt(poly,ix0*dx,iy0*dy);
  if ( cen<last && *(cen+2)==ix0-1 && *(cen+3)==iy0+1 )
  { iy0 += 2;  side = 3;  cen += 2; }
  else { ix0 -= 2; corner++; }
  ix1 = ix0;  iy1 = iy0;

  inc = 0;  /*If too many, found inf. loop bug*/

  /*Continue outline until we return to starting point*/
  while (cen!=vxy || side!=0 || corner!=0)
  { switch ( side )
    { case 0 :
        if ( cen<last && *(cen+2)==ix0-1 && *(cen+3)==iy0+1 ) { iy1 += 2;  side = 3;  cen += 2;  corner = 0; }
        else if ( cen>vxy && *(cen-2)==ix0-1 && *(cen-1)==iy0+1 ) { iy1 += 2;  side = 3;  cen -= 2;  corner = 0; }
        else if (corner==0) { ix1 -= 2;  corner++; }
        else if (corner==1) side++;
        else  announce_bug("vxy2rpoly(0)");
        break;
      case 1 :
        if ( cen<last && *(cen+2)==ix0-1 && *(cen+3)==iy0-1 ) { ix1 -= 2;  side = 0;  cen += 2;  corner = 1; }
        else if ( cen>vxy && *(cen-2)==ix0-1 && *(cen-1)==iy0-1 ) { ix1 -= 2;  side = 0;  cen -= 2;  corner = 1; }
        else if (corner==1) { iy1 -= 2;  corner++; }
        else if (corner==2) side++;
        else  announce_bug("vxy2rpoly(1)");
        break;
      case 2 :
        if ( cen<last && *(cen+2)==ix0+1 && *(cen+3)==iy0-1 ) { iy1 -= 2;  side = 1;  cen += 2;  corner = 2; }
        else if ( cen>vxy && *(cen-2)==ix0+1 && *(cen-1)==iy0-1 ) { iy1 -= 2;  side = 1;  cen -= 2;  corner = 2; }
        else if (corner==2) { ix1 += 2;  corner++; }
        else if (corner==3) side++;
        else  announce_bug("vxy2rpoly(2)");
        break;
      case 3 :
        if ( cen<last && *(cen+2)==ix0+1 && *(cen+3)==iy0+1 ) { ix1 += 2;  side = 2;  cen += 2;  corner = 3; }
        else if ( cen>vxy && *(cen-2)==ix0+1 && *(cen-1)==iy0+1 ) { ix1 += 2;  side = 2;  cen -= 2;  corner = 3; }
        else if (corner==3) { iy1 += 2;  corner = 0; }
        else if (corner==0) side = 0;
        else  announce_bug("vxy2rpoly(3)");
        break;
    } /*switch*/
    if ( (ix0!=ix1 || ix0!=ix0_old) && (iy0!=iy1 || iy0!=iy0_old) )
    { add_rpoly_pt(poly,ix0*dx,iy0*dy);
      ix0_old = ix0;  iy0_old = iy0;
    }
    ix0 = ix1;  iy0 = iy1;

    /*Check for inf. loop bug*/
    if ( (++inc) > 10*num )
      announce_bug("vxy2rpoly(4)");
  } /*while*/
  /*Add the last (same as 1st) pt in*/
  add_rpoly_pt(poly,ix0*dx,iy0*dy);
} /*vxy2rpoly*/

/*
**  The following 3 routines are used by xgeom:
**     init_vias, clear_vias, add_e_via
**
**  The rest of the via routines are in xdvia.c
**
*/

init_vias(vias) via_list *vias;
/*************************************************************/
/* Initialize the info in vias.                              */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* krs23Mar90 */
{ vias->n = 0;
  vias->via0 = NULL;
} /*init_vias*/

clear_vias(vias) via_list *vias;
/*************************************************************/
/* Close down all allocated memory in vias.                  */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* krs23Mar90 */
{ 
	if (vias->via0 != NULL) sFree((char *)vias->via0);
	vias->n = 0;  vias->via0 = NULL;
} /*clear_vias*/

add_e_via(poly_index,ipt,ilev,typ,vias,line_nu)  
int poly_index,ipt,ilev;
via_list *vias;
int typ,line_nu;
/*************************************************************/
/* Add an edge via. 					     */
/* If no via on the poly-th poly at the ipt-th pt, add.      */
/* If already exists returns -1, if added returns 0	     */
/* (C) COPYRIGHT 1986 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* krs27Mar90 */
{ 
	via_typ *iv;

	/*If overlapping via is already here, skip*/
	for (iv=vias->via0; iv<vias->via0+vias->n; iv++) {
		if (iv->ilev==ilev && iv->poly_index==poly_index && iv->ipt==ipt ) {
			return(-1);
		}
	}

	vias->n++;
	vias->via0 = (vias->via0 == NULL)
		? (via_typ *)new(sizeof(via_typ))
		: (via_typ *)rnew((char *)vias->via0,vias->n*sizeof(via_typ));
	iv = vias->via0 + vias->n - 1;
	iv->poly_index = poly_index;  iv->ipt = ipt;
	iv->ilev = ilev;
	iv->center_pts = NULL;
	iv->num_center_pts = 0;
	iv->ipoly = NULL;
	iv->subsection = 1;
	iv->line_nu = line_nu;
	iv->typ = typ;
	return(0);
} /* add_e_via */

cmp_yx(xy1,xy2) int *xy1,*xy2;
/*************************************************************/
/* Compare function for sorting xy points.                   */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr12may89 */ 
{ xy1++;  xy2++;
  if (*xy1 > *xy2) return(1);
  if (*xy1 < *xy2) return(-1);
  xy1--;  xy2--;
  if (*xy1 > *xy2) return(1);
  if (*xy1 < *xy2) return(-1);
  return(0);
} /*cmp_yx*/

cmp_xy(xy1,xy2) int *xy1,*xy2;
/*************************************************************/
/* Compare function for sorting xy points.                   */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr02may89 */ 
{ if (*xy1 > *xy2) return(1);
  if (*xy1 < *xy2) return(-1);
  xy1++;  xy2++;
  if (*xy1 > *xy2) return(1);
  if (*xy1 < *xy2) return(-1);
  return(0);
} /*cmp_xy*/

cmpxrect(xy1,xy2) int *xy1,*xy2;
/*************************************************************/
/* Compare function for sorting rectangles, x is m.s.d.      */
/* (C) COPYRIGHT 1990 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr08oct90 */ 
{ if ( *xy1 >= *(xy2+2) ) return(1);
  if ( *(xy1+2) <= *xy2 ) return(-1);
  if ( *(xy1+1) >= *(xy2+3) ) return(1);
  if ( *(xy1+3) <= *(xy2+1) ) return(-1);
  return(0);  /*One rect is equal to or contained in the other*/
} /*cmpxrect*/

cmpyrect(xy1,xy2) int *xy1,*xy2;
/*************************************************************/
/* Compare function for sorting rectangles, y is m.s.d.      */
/* (C) COPYRIGHT 1990 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr08oct90 */ 
{ if ( *(xy1+1) >= *(xy2+3) ) return(1);
  if ( *(xy1+3) <= *(xy2+1) ) return(-1);
  if ( *xy1 >= *(xy2+2) ) return(1);
  if ( *(xy1+2) <= *xy2 ) return(-1);
  return(0);  /*One rect is equal to or contained in the other*/
} /*cmpyrect*/

ck_rect(xy) int *xy;
/*************************************************************/
/* Check rectangle for valid parameters.                     */
/* (C) COPYRIGHT 1990 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr09oct90 */ 
{ int tmp;
  if (*xy == *(xy+2)) return(-1);
  if (*xy > *(xy+2)) { tmp = *xy;  *xy = *(xy+2);  *(xy+2) = tmp; }
  xy++;
  if (*xy == *(xy+2)) return(-1);
  if (*xy > *(xy+2)) { tmp = *xy;  *xy = *(xy+2);  *(xy+2) = tmp; }
  return(0);
} /*ck_rect*/

tri_fill(poly,lev,plev,Box)  Ipoly *poly;  msub_list *lev;  Poly_list *plev;  Box_info *Box;
/*************************************************************/
/* Fill diagonal edges of ipoly with triangle subsections.   */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/*Convert the polygon to a staircase too*/
/* jcr20may89 */ 
{ int ix,iy,idx,idy;
  int *x0p,*x1p;
  int xy[2];
  Ipoly *polyb,polyb_struct;

  if (poly->n<4) return;
  init_ipoly( polyb = &polyb_struct );

  for (x1p=(x0p=poly->xy0)+2; x1p<poly->xy0+2*poly->n; x0p += 2,x1p += 2)
  { add_ipoly_pt(polyb,*x0p,*(x0p+1));
    if (*x0p != *x1p && *(x0p+1) != *(x1p+1) )
    { idx = (*x1p > *x0p) ? 1 : -1;
      idy = (*(x1p+1) > *(x0p+1)) ? 1 : -1;
      for (ix = *x0p+idx, iy = *(x0p+1)+idy; (idx==1)?ix<*x1p:ix>*x1p; ix += 2*idx, iy += 2*idy)
      { xy[0] = ix;  xy[1] = iy;
        if ( (idx>0)==(idy>0) )  /*Diag is pos slope*/
	{ if (pt_in_ipoly(poly,ix-2,iy) && pt_in_ipoly(poly,ix,iy+2))
          { add_tri_ck(MTPP,MTNN,xy,Box,lev);
	    add_ipoly_pt(polyb,ix-1,iy+1);
            add_ipoly_pt(polyb,ix+idx,iy+idy);
          }
	  else if (pt_in_ipoly(poly,ix+2,iy) && pt_in_ipoly(poly,ix,iy-2))
          { add_tri_ck(MTNN,MTPP,xy,Box,lev);
	    add_ipoly_pt(polyb,ix+1,iy-1);
            add_ipoly_pt(polyb,ix+idx,iy+idy);
	  }
          else
	  { add_ipoly_pt(polyb,ix+1,iy-1);
            add_ipoly_pt(polyb,ix+idx,iy+idy);
          }
        }
        else /*Diag is neg slope*/
	{ if (pt_in_ipoly(poly,ix+2,iy) && pt_in_ipoly(poly,ix,iy+2))
          { add_tri_ck(MTNP,MTPN,xy,Box,lev);
	    add_ipoly_pt(polyb,ix+1,iy+1);
            add_ipoly_pt(polyb,ix+idx,iy+idy);
          }
	  else if (pt_in_ipoly(poly,ix-2,iy) && pt_in_ipoly(poly,ix,iy-2))
          { add_tri_ck(MTPN,MTNP,xy,Box,lev);
	    add_ipoly_pt(polyb,ix-1,iy-1);
            add_ipoly_pt(polyb,ix+idx,iy+idy);
          }
          else
	  { add_ipoly_pt(polyb,ix-1,iy-1);
            add_ipoly_pt(polyb,ix+idx,iy+idy);
          }
        }
      }
    }
  }
  add_ipoly_pt(polyb,*poly->xy0,*(poly->xy0+1));
  move_ipoly(poly,polyb);
  clear_ipoly(polyb);
} /*tri_fill*/

add_tri_ck(id_add,id_opp,tri_xy,Box,lev)  int id_add,id_opp,tri_xy[2];
					  Box_info *Box;  msub_list *lev;
/*************************************************************/
/* Add the triangle as long as id_opp tri not at same x,y.   */
/* (C) COPYRIGHT 1993 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr29jun93 */ 
{ int xy[6];
  msub *im;

  /*Search thru all subs for the opposite triangle*/
  for (im=lev->msub0; im<lev->msub0+lev->n; im++)
    if ( im->id == id_opp && tri_xy[0] == *(im->ip) && tri_xy[1] == *(im->ip+1) )
    { /*remove the offending triangle*/
      remove_msub(im,lev);
      /*add some subs to cover both the id_add and id_opp subs*/
      xy[0] = tri_xy[0] - 3;  xy[1] = tri_xy[1] - 1;
      xy[2] = tri_xy[0] + 1;  xy[3] = tri_xy[1] + 1;
      xy[4] = tri_xy[0] - 1;  xy[5] = 0;
      add_msub(MXR,xy,Box,lev);
      xy[0] = tri_xy[0] - 1;  xy[1] = tri_xy[1] - 1;
      xy[2] = tri_xy[0] + 3;  xy[3] = tri_xy[1] + 1;
      xy[4] = tri_xy[0] + 1;  xy[5] = 0;
      add_msub(MXR,xy,Box,lev);
      xy[0] = tri_xy[0] - 1;  xy[1] = tri_xy[1] - 3;
      xy[2] = tri_xy[0] + 1;  xy[3] = tri_xy[1] + 1;
      xy[4] = 0;              xy[5] = tri_xy[1] - 1;
      add_msub(MYR,xy,Box,lev);
      xy[0] = tri_xy[0] - 1;  xy[1] = tri_xy[1] - 1;
      xy[2] = tri_xy[0] + 1;  xy[3] = tri_xy[1] + 3;
      xy[4] = 0;              xy[5] = tri_xy[1] + 1;
      add_msub(MYR,xy,Box,lev);
      return;
    }

  /*Add the desired triangle*/
  add_msub(id_add,tri_xy,Box,lev);
  return;
} /*add_tri_ck*/

cor_fill(poly,lev,plev,Box,xlim,ylim) Ipoly *poly;  msub_list *lev;
                       Poly_list *plev;  Box_info *Box;  int xlim,ylim;
/*************************************************************/
/* Add end subs to each right angle corner.                  */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr20may89 */ 
{ int *x0p,*x1p,*x2p;
  int dx0,dy0,dx1,dy1,dx2,dy2;
  int sx0,sy0,sx1,sy1;
  int xy[2];
  int i,ilim,inside,good_corner;

  if (poly->n<4) return;

  x0p = poly->xy0;  x1p = x0p + 2;  x2p = x1p + 2;
  do { if (x2p>=poly->xy0+2*poly->n-2) x2p = poly->xy0;  /*Pick up 1st corner*/
       
       /*Length of each line segment*/
       dx0 = abs( *x1p - *x0p );  dy0 = abs( *(x1p+1) - *(x0p+1) ); /*1st seg*/
       dx1 = abs( *x2p - *x1p );  dy1 = abs( *(x2p+1) - *(x1p+1) ); /*2nd seg*/

       /*Get direction from middle point of each edge*/
       sx0 = sy0 = sx1 = sy1 = 0;
       if (dx0) sx0 = (*x1p > *x0p) ? -1 : 1;
       if (dy0) sy0 = (*(x1p+1) > *(x0p+1)) ? -1 : 1;
       if (dx1) sx1 = (*x1p > *x2p) ? -1 : 1;
       if (dy1) sy1 = (*(x1p+1) > *(x2p+1)) ? -1 : 1;

       /*Check for inside or outside corner...
              outside  ________               
              corner   |      |
                       |      |
              _________|      |________
                     inside corners
              _________________________
       */
       inside = pt_in_ipoly(poly,*x1p-sx0,*(x1p+1)-sy0);

       good_corner = dx0+dy0>2 && dx1+dy1>2;    /*Both line segments longer than 2*/
       if ( !good_corner )         /*Check for one sub wide stub*/
       { dx2 = abs( *(x2p+2) - *x2p );  dy2 = abs( *(x2p+3) - *(x2p+1) ); /*3rd seg*/
	 /*Make sure is end of stub, not staircased diagonal edge*/
	 if ( dx1 == 0 && dx0 != 0 && dx2 != 0 && (dx0>0) != (dx2>0) )
	   good_corner = 1;
	 else if ( dy1 == 0 && dy0 != 0 && dy2 != 0 && (dy0>0) != (dy2>0) )
	   good_corner = 1;
       }

       if (!inside &&
           *x1p!=0 && *x1p!=Box->Capm &&   /*Middle pt not on box edge*/
	   *(x1p+1)!=0 && *(x1p+1)!=Box->Capn &&
           (dx0==0)!=(dx1==0) &&
	   (dy0==0)!=(dy1==0) &&           /*Is a right angle corner*/
           !another_ipoly(*x1p,*(x1p+1),poly,plev) )  /*No other poly in area*/
       { /*add end subs from x1 back towards x0*/
         /*do xlim subs or half way along line, which ever is less*/
	 ilim = (dx0+dy0+2)/4;
	 if (ilim>xlim && dy0==0) ilim = xlim;
	 if (ilim>ylim && dx0==0) ilim = ylim;
	 xy[0] = *x1p + sx0;  xy[1] = *(x1p+1) + sy0;
         for (i=0; i<ilim; i++,xy[0] += 2*sx0, xy[1] += 2*sy0)
	 { if (sy0==0) /*Horizontal line*/
	   { if ( pt_in_ipoly(poly,xy[0],xy[1]+1) && another_ipoly(xy[0],xy[1]-5,NULL,plev) )
               add_msub(MENDYN,xy,Box,lev);
	     else if (pt_in_ipoly(poly,xy[0],xy[1]-1) && another_ipoly(xy[0],xy[1]+5,NULL,plev) )
               add_msub(MENDYP,xy,Box,lev);
             else break;
           }
	   else
	   { if (pt_in_ipoly(poly,xy[0]+1,xy[1]) && another_ipoly(xy[0]-5,xy[1],NULL,plev) )
               add_msub(MENDXN,xy,Box,lev);
	     else if (pt_in_ipoly(poly,xy[0]-1,xy[1]) && another_ipoly(xy[0]+5,xy[1],NULL,plev) )
               add_msub(MENDXP,xy,Box,lev);
             else break;
           }
         }
	 
         /*add end subs from x1 forward towards x2*/
         /*do xlim subs or half way along line, which ever is less*/
	 ilim = (dx1+dy1+2)/4;
	 if (ilim>xlim && dy1==0) ilim = xlim;
	 if (ilim>ylim && dx1==0) ilim = ylim;
	 xy[0] = *x1p + sx1;  xy[1] = *(x1p+1) + sy1;
	 for (i=0; i<ilim; i++,xy[0] += 2*sx1, xy[1] += 2*sy1)
	 { if (sy1==0) /*Horizontal line*/
	   { if (pt_in_ipoly(poly,xy[0],xy[1]+1) && another_ipoly(xy[0],xy[1]-5,NULL,plev) )
               add_msub(MENDYN,xy,Box,lev);
	     else if (pt_in_ipoly(poly,xy[0],xy[1]-1) && another_ipoly(xy[0],xy[1]+5,NULL,plev) )
               add_msub(MENDYP,xy,Box,lev);
             else break;
           }
	   else
	   { if (pt_in_ipoly(poly,xy[0]+1,xy[1]) && another_ipoly(xy[0]-5,xy[1],NULL,plev) ) 
               add_msub(MENDXN,xy,Box,lev);
	     else if (pt_in_ipoly(poly,xy[0]-1,xy[1]) && another_ipoly(xy[0]+5,xy[1],NULL,plev) )
               add_msub(MENDXP,xy,Box,lev);
             else break;
           }
         }
       }
       x0p = x1p;  x1p = x2p;  x2p += 2;
     } while (x0p != poly->xy0);
} /*cor_fill*/

ipoly_rm_acute(poly) Ipoly *poly;
/*************************************************************/
/* Remove all acute internal angles.                         */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr21may89 */ 
{ int *x0p,*x1p,*x2p;
  int dx0,dy0,dx1,dy1;
  Ipoly *polyb,polyb_struct;
  int change=1;

  while (change)
  { if (poly->n<4) { clear_ipoly(poly);  return; }
    init_ipoly( polyb = &polyb_struct );
    change = 0;

    x0p = poly->xy0;  x1p = x0p + 2;  x2p = x1p + 2;
    do { if (x2p>=poly->xy0+2*poly->n-2) x2p = poly->xy0;  /*Pick up first corner*/
         if (!change)  /*Remove only one corner per while(change) loop (can't handle 2 conseq. acute corners)*/
         { dx0 = *x1p - *x0p;  if (dx0>0) dx0 = 1;  if (dx0<0) dx0 = -1;
           dx1 = *x1p - *x2p;  if (dx1>0) dx1 = 1;  if (dx1<0) dx1 = -1;
           dy0 = *(x1p+1) - *(x0p+1);  if (dy0>0) dy0 = 1;  if (dy0<0) dy0 = -1;
           dy1 = *(x1p+1) - *(x2p+1);  if (dy1>0) dy1 = 1;  if (dy1<0) dy1 = -1;
         }
         if ( !change && (dx0*dx1==1 || dy0*dy1==1) &&
  	      !pt_in_ipoly(poly,*x1p+dx0,*(x1p+1)+dy0) )  /*which side is metal on*/
         { /*Found an acute interior angle*/
           add_ipoly_pt(polyb,*x1p-2*dx0,*(x1p+1)-2*dy0);
	   add_ipoly_pt(polyb,*x1p-2*dx1,*(x1p+1)-2*dy1);
	   change = 1;
         }
         else add_ipoly_pt(polyb,*x1p,*(x1p+1));
         x0p = x1p;  x1p = x2p;  x2p += 2;
       } while (x0p != poly->xy0);
    add_ipoly_pt(polyb,*polyb->xy0,*(polyb->xy0+1));
    move_ipoly(poly,polyb);
  }
} /*ipoly_rm_acute*/

another_ipoly(ix,iy,ipoly,plev) int ix,iy;  Ipoly *ipoly;  Poly_list *plev;
/*************************************************************/
/* Return true if another ipoly (not the one passed) at x,y  */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr30aug89 */ 
{ Ipoly *jpoly;

  for (jpoly=plev->ipoly0; jpoly<plev->ipoly0+plev->n; jpoly++)
    if ( jpoly != ipoly &&
         ( pt_in_ipoly(jpoly,ix+1,iy) || pt_in_ipoly(jpoly,ix-1,iy) ||
           pt_in_ipoly(jpoly,ix,iy+1) || pt_in_ipoly(jpoly,ix,iy-1) ) ) return(1);

  return(0);
} /*another_ipoly*/

another_ipoly2(ix,iy,ipoly,plev) int ix,iy;  Ipoly *ipoly;  Poly_list *plev;
/*************************************************************/
/* Return true if another ipoly (not the one passed) at x,y  */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr23dec92 -- just like another_ipoly, except looks only at ix,iy */
{ Ipoly *jpoly;

  for (jpoly=plev->ipoly0; jpoly<plev->ipoly0+plev->n; jpoly++)
    if ( jpoly != ipoly && pt_in_ipoly(jpoly,ix,iy) )
      return(1);

  return(0);
} /*another_ipoly*/

/*Globals for routines to check vertices*/
static Ipoly *ckp[3];
static Ipoly *ckp_stop[3];
static int ref0,ref1,ref2,ref3;

init_ckvert(plev0,ilev,nlev,Box) Poly_list *plev0; int ilev,nlev;  Box_info *Box;
/*************************************************************/
/* Initialize ckx/yvert for next polygon.                    */
/* (C) COPYRIGHT 1992 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr04feb92 */
{ Poly_list *plev;

  if (ilev<0 || ilev>nlev-2) /*ilev should not include ground*/
    announce_bug("init_ckvert(0)");

  /*Polygon level presently being subsectioned*/
  plev = plev0 + ilev;
  ckp[0] = plev->ipoly0;
  ckp_stop[0] = ckp[0] + plev->n;

  if ( !Box->ckp_lev )  /*Don't check up or down*/
  { ckp[1] = NULL;  ckp_stop[1] = NULL;
    ckp[2] = NULL;  ckp_stop[2] = NULL;
    return;
  }

  /*Ck for polygon list previously subsectioned*/
  if ( ilev > 0 )
  { plev = plev0 + ilev - 1;
    ckp[1] = plev->ipoly0;
    ckp_stop[1] = ckp[1] + plev->n;
  }
  else
  {  ckp[1] = NULL;  ckp_stop[1] = NULL; }

  /*Set up polygon list to be subsectined next*/
  plev = plev0 + ilev + 1;  /*include ground ck*/
  ckp[2] = plev->ipoly0;
  ckp_stop[2] = ckp[2] + plev->n;

  ref0 = Box->ref[0];  ref1 = Box->Capm - Box->ref[1];
  ref2 = Box->ref[2];  ref3 = Box->Capn - Box->ref[3];
} /*init_ckvert*/

ckxvert( x,y,dy )  int x,y,dy;
/*************************************************************/
/* Return true if a vertice or vert. edge at (x,y+0/+dy).    */
/* (C) COPYRIGHT 1992 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr04feb92 */ 
{ int *xp,*yp,*xstop;
  Ipoly *ipoly;
  int i;

  /*Reference planes are considered vertices too*/
  if (x==ref0 || x==ref1) return(1);

  dy = y + dy;  /*dy is now a coord, not a delta*/
  if ( dy > y )  /*make sure y > dy*/
  { i = y;  y = dy;  dy = i; }

  /*Check all polygons on 3 check levels*/
  for (i=0; i<3; i++)
    for (ipoly=ckp[i]; ipoly<ckp_stop[i]; ipoly++) if (ipoly->n != 0)
    { xstop = ipoly->xy0 + 2 * ipoly->n - 2;
      yp = ipoly->xy0;  xp = yp++;
      for ( ; xp<xstop; yp++,xp=yp++)
        if ( *xp == x && *(xp+2) == x )               /*Vert. line seg.*/
        { if ( ( (*yp > y) != (*(yp+2) > y) )   ||    /*poly edge crosses y*/
               ( (*yp > dy) != (*(yp+2) > dy) ) ||    /*poly edge crosses dy*/
	       ( (*yp >= dy) && (*yp <= y) )    ||    /*1st pt within y..dy*/
	       ( (*(yp+2) >= dy) && (*(yp+2) <= y) )  /*2nd pt within y..dy*/
             ) return(1);
        }
    }

  return(0);
} /*ckxvert*/

ckyvert( x,y,dx )  int x,y,dx;
/*************************************************************/
/* Return true if a vertice or horz. edge at (x+0/+dx,y).    */
/* (C) COPYRIGHT 1992 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* jcr04feb92 */ 
{ int *xp,*yp,*xstop;
  Ipoly *ipoly;
  int i;

  /*Reference planes are considered vertices too*/
  if (y==ref2 || y==ref3) return(1);

  dx = x + dx;  /*dx is now a coord, not a delta*/
  if ( dx > x )  /*make sure x > dx*/
  { i = x;  x = dx;  dx = i; }

  /*Check all polygons on 3 check levels*/
  for (i=0; i<3; i++)
    for (ipoly=ckp[i]; ipoly<ckp_stop[i]; ipoly++) if (ipoly->n != 0)
    { xstop = ipoly->xy0 + 2 * ipoly->n - 2;
      yp = ipoly->xy0;  xp = yp++;
      for ( ; xp<xstop; yp++,xp=yp++)
        if ( *yp == y && *(yp+2) == y )               /*horiz. line seg.*/
          if ( ( (*xp > x) != (*(xp+2) > x) )   ||    /*poly edge crosses x*/
               ( (*xp > dx) != (*(xp+2) > dx) ) ||    /*poly edge crosses dx*/
	       ( (*xp >= dx) && (*xp <= x) )    ||    /*1st pt within x..dx*/
	       ( (*(xp+2) >= dx) && (*(xp+2) <= x) )  /*2nd pt within x..dx*/
             ) return(1);
    }

  return(0);
} /*ckyvert*/

Rpoly * 
copy_rpoly(from)  Rpoly *from;
/*************************************************************/
/* Do the indicated copy.                                    */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* krs12feb92 */ 
{ 
  int i;
  Rpoly *to;

  to = (Rpoly *) new( sizeof(Rpoly) ); 

  if (from==NULL)
    announce_bug("copy_rpoly");
  to->n = from->n;
  to->max = from->max;
  to->mode = from->mode;
  to->icond = from->icond;
  to->xmin = from->xmin;
  to->ymin = from->ymin;
  to->xmax = from->xmax;
  to->ymax = from->ymax;
  to->manhatten = from->manhatten;
  to->unsel = from->unsel;  
  to->id = from->id;
  to->xgeom_id = from->xgeom_id;

  to->xy0 = (Rtyp *)new( 2*to->max*sizeof(Rtyp) );
  for (i=0; i<= to->n*2; ++i ) {
	to->xy0[i] = from->xy0[i];
  }
  return (to);

} /*copy_rpoly*/

dump_rpoly(rpoly,dx,dy)  Rpoly *rpoly;  Rtyp dx,dy;
/*************************************************************/
/* Could not convert a rpoly to ipoly, print out coordinates */
/* for debug purposes					     */
/* (C) COPYRIGHT 1989 JAMES C. RAUTIO.  ALL RIGHTS RESERVED. */
/*************************************************************/
/* krs6jun93 */ 
{ 
  int i;
  fprintf(tracefile,"Trace for poly_rtoi\n");
  fprintf(tracefile,"dx = %f, dy = %f\n");
  fprintf(tracefile,"rpoly->n = %d\n",rpoly->n);
  fprintf(tracefile,"rpoly->mode = %d\n",rpoly->mode);
  for (i=0; i<rpoly->n; ++i ) {
    fprintf(tracefile,"%f, %f\n",rpoly->xy0[2*i], rpoly->xy0[2*i+1]);

  }
}

debug_polyutil(tf)
FILE *tf;
{
  tracefile = tf;
  debug_poly = TRUE;
}
